# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""std_calibration state machine v2.0

Revision ID: 0b4046b7c010
Revises: 4354741bf742
Create Date: 2021-12-09 15:03:20.568128

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "0b4046b7c010"
down_revision = "4354741bf742"
branch_labels = None
depends_on = None

"""
Change std_calibration state machine from its current form:
[Start] - <capability-submitted> {QueueWorkflow std_calibration} -> [Queued]
[Queued] - <resume> {ExecuteWorkflow std_calibration} -> [Executing]
[Executing] - <workflow-complete> {SendMessage execution_complete} {SendNotification workflow-complete} {AnnounceQa} -> [Awaiting QA]
[Executing] - <workflow-failed> {SendMessage execution_failed} {SendNotification workflow-failed} -> [Failed]
[Awaiting QA] - <qa-pass> {QaPass} -> [Complete]
[Awaiting QA] - <qa-fail> {SendMessage execution_failed} -> [Failed]

to its new form:
[Start] - <capability-submitted> {QueueWorkflow std_calibration} -> [Queued]
[Queued] - <resume> {ExecuteWorkflow std_calibration} -> [Executing]
[Executing] - <workflow-complete> {AnnounceQa} -> [Awaiting QA]
[Executing] - <workflow-failed> {SendMessage execution_failed} {SendNotification workflow-failed} -> [Failed]
[Awaiting QA] - <qa-pass> {QaPass} -> [Ingesting]
[Awaiting QA] - <qa-fail> {QaFail} {SendMessage execution_failed} -> [Failed]
[Ingesting] - <workflow-complete> {SendMessage execution_complete} {SendNotification workflow-complete} -> [Complete]
[Ingesting] - <workflow-failed> {SendMessage execution_failed} {SendNotification ingest-failed} -> [Failed]
"""


def upgrade():
    # Create new state: Ingesting
    op.execute("INSERT INTO capability_states (capability_name, state) VALUES ('std_calibration', 'Ingesting')")

    # Modify existing transition:
    # [Awaiting QA] - <qa-pass> {QaPass} -> [Complete]
    # turns into
    # [Awaiting QA] - <qa-pass> {QaPass} -> [Ingesting]
    ingesting_instead_of_complete_sql = """
    UPDATE capability_state_transitions SET to_state = 'Ingesting'
    WHERE capability_name = 'std_calibration'
        AND from_state = 'Awaiting QA'
        AND pattern = 'type == qa-pass';
    """
    op.execute(ingesting_instead_of_complete_sql)

    # Delete existing actions:
    # [Executing] - <workflow-complete> {SendMessage execution_complete} {SendNotification workflow-complete} {AnnounceQa} -> [Awaiting QA]
    # goes to
    # [Executing] - <workflow-complete> {AnnounceQa} -> [Awaiting QA]
    delete_extra_qa_pass_actions_sql = """
    DELETE FROM capability_state_actions
    WHERE action_type <> 'AnnounceQa'
      AND transition_id = (SELECT transition_id
                           FROM capability_state_transitions
                           WHERE capability_name = 'std_calibration'
                               AND pattern = 'type == workflow-complete'
                               AND from_state = 'Executing'
                               AND to_state = 'Awaiting QA'
    );
    """
    op.execute(delete_extra_qa_pass_actions_sql)

    # Add new transitions/actions
    # [Ingesting] - <workflow-complete> {SendMessage execution_complete} {SendNotification workflow-complete} -> [Complete]
    # [Ingesting] - <workflow-failed> {SendMessage execution_failed} {SendNotification ingest-failed} -> [Failed]
    add_new_ingesting_transitions_sql = """
    INSERT INTO capability_state_transitions (capability_name, from_state, pattern, to_state)
    VALUES ('std_calibration', 'Ingesting', 'type == workflow-complete', 'Complete');
    INSERT INTO capability_state_transitions (capability_name, from_state, pattern, to_state)
    VALUES ('std_calibration', 'Ingesting', 'type == workflow-failed', 'Failed');
    """
    op.execute(add_new_ingesting_transitions_sql)

    add_new_ingesting_actions_sql = """
    INSERT INTO capability_state_actions (transition_id, action_type, arguments)
    SELECT transition_id, 'SendMessage', 'execution_complete'
    FROM capability_state_transitions
    WHERE capability_name = 'std_calibration'
        AND from_state = 'Ingesting'
        AND pattern = 'type == workflow-complete';
    INSERT INTO capability_state_actions (transition_id, action_type, arguments)
    SELECT transition_id, 'SendNotification', 'workflow-complete'
    FROM capability_state_transitions
    WHERE capability_name = 'std_calibration'
      AND from_state = 'Ingesting'
      AND pattern = 'type == workflow-complete';
    INSERT INTO capability_state_actions (transition_id, action_type, arguments)
    SELECT transition_id, 'SendMessage', 'execution_failed'
    FROM capability_state_transitions
    WHERE capability_name = 'std_calibration'
      AND from_state = 'Ingesting'
      AND pattern = 'type == workflow-failed';
    INSERT INTO capability_state_actions (transition_id, action_type, arguments)
    SELECT transition_id, 'SendNotification', 'workflow-failed'
    FROM capability_state_transitions
    WHERE capability_name = 'std_calibration'
      AND from_state = 'Ingesting'
      AND pattern = 'type == workflow-failed';
    """
    op.execute(add_new_ingesting_actions_sql)

    # Add new action to transition:
    # [Awaiting QA] - <qa-fail> {SendMessage execution_failed} -> [Failed]
    # turns into
    # [Awaiting QA] - <qa-fail> {QaFail} {SendMessage execution_failed} -> [Failed]
    add_new_qa_fail_action_sql = """
    INSERT INTO capability_state_actions (transition_id, action_type)
    SELECT transition_id, 'QaFail'
    FROM capability_state_transitions
    WHERE capability_name = 'std_calibration'
      AND from_state = 'Awaiting QA'
      AND pattern = 'type == qa-fail';
    """
    op.execute(add_new_qa_fail_action_sql)


def downgrade():
    # Delete new QaFail action
    delete_qa_fail_action_sql = """
    DELETE FROM capability_state_actions
    WHERE action_type = 'QaFail'
        AND transition_id = (SELECT transition_id
                           FROM capability_state_transitions
                           WHERE capability_name = 'std_calibration'
                             AND from_state = 'Awaiting QA'
                             AND pattern = 'type == qa-fail'
    );
    """
    op.execute(delete_qa_fail_action_sql)

    # Delete new Ingesting actions
    delete_ingesting_actions_sql = """
    DELETE FROM capability_state_actions
    WHERE transition_id = (SELECT transition_id
                           FROM capability_state_transitions
                           WHERE capability_name = 'std_calibration'
                             AND from_state = 'Ingesting'
                             AND to_state = 'Complete'
    );
    DELETE FROM capability_state_actions
    WHERE transition_id = (SELECT transition_id
                           FROM capability_state_transitions
                           WHERE capability_name = 'std_calibration'
                             AND from_state = 'Ingesting'
                             AND to_state = 'Failed'
    );
    """
    op.execute(delete_ingesting_actions_sql)

    # Delete new Ingesting transitions
    delete_ingesting_transitions_sql = """
    DELETE FROM capability_state_transitions WHERE capability_name = 'std_calibration' AND from_state = 'Ingesting';
    """
    op.execute(delete_ingesting_transitions_sql)

    # Add back extra qa-pass actions
    add_back_qa_pass_actions_sql = """
    INSERT INTO capability_state_actions (transition_id, action_type, arguments)
    SELECT transition_id, 'SendMessage', 'execution_complete'
    FROM capability_state_transitions
    WHERE capability_name = 'std_calibration'
      AND from_state = 'Executing'
      AND pattern = 'type == workflow-complete';
    INSERT INTO capability_state_actions (transition_id, action_type, arguments)
    SELECT transition_id, 'SendNotification', 'workflow-complete'
    FROM capability_state_transitions
    WHERE capability_name = 'std_calibration'
      AND from_state = 'Executing'
      AND pattern = 'type == workflow-complete';
    """
    op.execute(add_back_qa_pass_actions_sql)

    # Revert [Awaiting QA] -> [Complete]
    back_to_complete_sql = """
    UPDATE capability_state_transitions SET to_state = 'Complete'
    WHERE capability_name = 'std_calibration'
      AND from_state = 'Awaiting QA'
      AND pattern = 'type == qa-pass';
    """
    op.execute(back_to_complete_sql)

    # Delete Ingesting state
    delete_ingesting_sql = """
    DELETE FROM capability_states
    WHERE capability_name = 'std_calibration'
        AND state = 'Ingesting';
    """
    op.execute(delete_ingesting_sql)
